#include "PDFDoc.h"
#include "OutputDev.h"
#include "GlobalParams.h"
#include "GfxState.h"
#include "GfxFont.h"

#include <string>
#include <cstdio>
#include <cstring>
#include <map>

static void UTF8Put(unsigned ch)
{
    const unsigned n = ch;
    if(n < 0x80)                 // <=7 bits
        std::putchar( (char)n );
    else
    {
        if(n < 0x800)            // <=11 bits
            std::putchar( (char)(0xC0 + (n>>6)) );
        else
        {
            if(n < 0x10000)      // <=16 bits
                std::putchar( (char)(0xE0 + (n>>12)) );
            else                 // <=21 bits
            {
                std::putchar( (char)(0xF0 + (n>>18)) );
                std::putchar( (char)(0x80 + ((n>>12)&63)) );
            }
            std::putchar( (char)(0x80 + ((n>>6)&63)) );
        }
        std::putchar( (char)(0x80 + (n&63)) );
    }
}

#define msg(...)

struct FontProps
{
    std::string name;
    std::string id;
    
    bool italic;
    bool oblique;
    bool bold;
    bool monospace;
    bool serif;
};

static FontProps getFontProps(GfxFont*font)
{
    Ref *ref = font->getID();
    GString *gstr = font->getName();
    char* fname = gstr==0 ? 0 : gstr->getCString();
    char buf[4096];
    if(fname==0) {
        sprintf(buf, "font-%d-%d", ref->num, ref->gen);
    } else {
        sprintf(buf, "%s-%d-%d", fname, ref->num, ref->gen);
    }
    
    FontProps props;
    props.id = buf;
    
    if(fname)
    {
        const char* plus = std::strchr(fname, '+');
        props.name = plus ? plus+1 : fname;
    }    
    
    props.italic  = font->isItalic();
    props.oblique = false;
    props.bold    = font->isBold();
    props.monospace=font->isFixedWidth();
    props.serif   = font->isSerif();
    
    #define FindAndDestroy(keyword, settings) \
      { std::string::size_type s = props.name.find(keyword); \
        if(s != props.name.npos) \
        { \
            settings; \
            props.name.erase(s, props.name.size()); \
      } }
    
    FindAndDestroy("-BoldOblique", props.bold=props.oblique=true);
    FindAndDestroy("-BoldItalic", props.bold=props.italic=true);
    FindAndDestroy("-Oblique", props.oblique=true);
    FindAndDestroy("-Italic", props.italic=true);
    FindAndDestroy("-Bold", props.bold=true);
    FindAndDestroy(",BoldOblique", props.bold=props.oblique=true);
    FindAndDestroy(",BoldItalic", props.bold=props.italic=true);
    FindAndDestroy(",Oblique", props.oblique=true);
    FindAndDestroy(",Italic", props.italic=true);
    FindAndDestroy(",Bold", props.bold=true);
    
    // TODO: -Regular, -Semibold, -It
    
    #undef FindAndDestroy
    
    return props;
}

static const double DefaultY = 1000;
class OutDev: public OutputDev
{
private:
    double LastY;
    std::string style;
    
    void SwitchStyle(const std::string& f)
    {
        if(f == style)return;
        if(!style.empty()) printf("</span>");
        
        if(!f.empty())
        {
            static std::map<std::string, unsigned> classmap;
            std::map<std::string, unsigned>::const_iterator i = classmap.find(f);
            unsigned clsid = 0;
            if(i == classmap.end())
            {
                clsid = classmap.size() + 1;
                classmap[f] = clsid;
                printf("\n<style>.c%u{%s}</style>\n", clsid, f.c_str());
            }
            else
            {
                clsid = i->second;
            }
            printf("<span class=c%u>", clsid);
        }
        style=f;
    }
public:
    OutDev(): OutputDev()
    {
        LastY=DefaultY;
        style="";
    }
    ~OutDev()
    {
        SwitchStyle("");
    }

    virtual GBool upsideDown() {return gFalse;}
    virtual GBool useDrawChar() {return gTrue;}
    virtual GBool useGradients() {return gFalse;}
    virtual GBool interpretType3Chars() {return gFalse;}
    virtual GBool needNonText() { return gFalse; }

   
    virtual void updateFont(GfxState *state) 
    {
        GfxFont*font = state->getFont();
        if(!font) return;
        
        FontProps props = getFontProps(font);
        
        std::string style = "font-family:" + props.name;
        
        if(props.monospace) style += ",monospace";
        if(props.serif) style += ",serif";
        else if(!props.monospace) style += ",sans-serif";
        // sans-serif is default in <body>, so only offer
        // it if no other last-resort is available
        
        if(props.italic) style += ";font-style:italic";
        if(props.oblique) style += ";font-style:oblique";
        if(props.bold) style += ";text-emphasis:bold";
        
        char Buf[4096];
        sprintf(Buf, ";font-size:%gpt", state->getTransformedFontSize());
        style += Buf;
        
        SwitchStyle(style);
    }
    
    //----- text drawing
    virtual void beginString(GfxState *state, GString *s)
    {
        msg("<debug> beginString '%s'\n", s->getCString());
    }
    virtual void endString(GfxState *state)
    {
        msg("<debug> endString\n");
        std::fflush(stdout);
    }
    virtual void drawChar(GfxState *state, double x, double y,
              double dx, double dy,
              double originX, double originY,
              CharCode code, int nBytes, Unicode *u, int uLen)
    {
        msg("<debug> drawChar %.1f, %.1f; %.1f, %.1f; %.1f, %.1f; code=%u (%c) (%u),bytes=%d,len=%d\n",
            x,y, dx,dy, originX,originY,
            code,*u, *u,
            nBytes,uLen);
        
        if(y != LastY)
        {
            double diff = LastY-y;
            
            if(diff < 25) printf("<br />\n");
            else
            {
                std::string st = style;
                SwitchStyle("");
                printf("\n<p />\n\n");
                SwitchStyle(st);
            }
            
            //printf(" {diff %g}\n", diff);
            LastY = y;
        }
        
        /* TODO: Check X difference too for indentation and spaces */
        
        UTF8Put(*u);
    }
    virtual void startPage(int pageNum, GfxState *state, double x1,double y1,double x2,double y2)
    {
        //printf("\n\n----\n\n");
        LastY=DefaultY;
        std::fflush(stdout);
    }
};

int main(int argc, const char*const* argv)
{
    if(argc < 2)
    {
        printf("%s", 
            "PDF to HTML converter - http://bisqwit.iki.fi/source/pdf2simplehtml.html\n"
            "Copyright (C) 1992,2006 Joel Yliluoma (http://iki.fi/bisqwit/)\n"
            "Usage: prog pdffilename > htmlfilename\n"
            "\n"
            "Note: The program only converts text (not images), and does\n"
            "not concern itself much with the layout.\n"
        );
        return 0;
    }

    {char p[]=""; globalParams = new GlobalParams(p);}
    
    GString tmp(argv[1]);
    
    PDFDoc* pdf = new PDFDoc(&tmp);
    std::fprintf(stderr, "%d pages\n", pdf->getNumPages());
    
    std::printf("<meta http-equiv='Content-type' content='text/html; charset=utf-8' />");
    std::printf("<meta name='generator' value='pdf2simplehtml by Bisqwit' />");
    std::printf("<style>body{font-family:sans-serif}</style>");
    std::printf("<body lang='ja_JP'>");

    {
     OutDev out;
     pdf->displayPages(&out, 1,pdf->getNumPages(),
                      72,72, 0,
                      true, true, 0);
 
    }
    std::printf("</body>");

    std::fprintf(stderr, "done\n");
    
    return 0;
}
